/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: esptest.c 2416 2018-08-25 15:30:39Z roi $
 */

/* 
 *  ESP-WROOM-02ƥȤ
 *  ɾΥץå她åESP-WROOM-02
 *  ꥻåȥåǤꥻåȸ塢ATޥɤȤäơ
 *  ESP-WROOM-02̿Ǥޤ
 *  㤨СAT+GMRϡAT GMR(return)ESP-WROOM-02
 *  ESP-WROOM-02μǡϥ󥽡ɽޤ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <target_syssvc.h>
#include <stdio.h>
#include <string.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "monitor.h"
#include "esptest.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

static char    aTxBuffer[TX_BUF_SIZE];
static uint8_t aRxBuffer[RX_BUF_SIZE];

int     rx_mode = MODE_DEFAULT;
queue_t rx_queue;

/*
 *  ǥХޥֹ
 */
#define BASE_CMD_LEN     4

static int_t mode_func(int argc, char **argv);
static int_t count_func(int argc, char **argv);
static int_t at_func(int argc, char **argv);


/*
 *  ǥХޥɥơ֥
 */
static const COMMAND_INFO at_command_info[] = {
	{"MODE",	mode_func},
	{"TCNT",	count_func},
	{"",		at_func}
};

#define NUM_DEVICE_CMD   (sizeof(at_command_info)/sizeof(COMMAND_INFO))

static const char at_name[] = "AT";
static const char at_help[] =
"  AT      MODE (no) echo mode\n"
"          TCNT      recived count\n"
"          any command\n";

static COMMAND_LINK at_command_link = {
	NULL,
	NUM_DEVICE_CMD,
	at_name,
	NULL,
	at_help,
	&at_command_info[0]
};

static int a2i(char *str)
{
	int num = 0;

	while(*str >= '0' && *str <= '9'){
		num = num * 10 + *str++ - '0';
	}
	return num;
}

/*
 *  ATޥؿ
 */
void at_info_init(intptr_t exinf)
{
	setup_command(&at_command_link);
}

/*
 *  ECHO⡼ؿ
 */
static int_t mode_func(int argc, char **argv)
{
	int mode = rx_mode;

	if(argc >= 2){
		mode = a2i(argv[1]);
		if(mode < 0 || mode > MODE_ECHO_HEX)
			mode = MODE_DEFAULT;
		rx_mode = mode;
	}
	printf("AT ECHO MODE(%d)\n", mode);
	return mode;
}

/*
 *  ȥޥɼؿ
 */
static int_t count_func(int argc, char **argv)
{
	printf("AT RECIVED COUNT(%d)(%d)\n", rx_queue.size, rx_queue.clen);
	return rx_queue.clen;
}


/*
 *  ATꥳޥɴؿ
 */
static int_t at_func(int argc, char **argv)
{
	int  i, arg_count = BASE_CMD_LEN;
	char *p = aTxBuffer;
	char *s, c;
	ER_UINT result;
	int  caps = 1;

	for(i = 0 ; i < argc ; i++){
		arg_count++;
		arg_count += strlen(argv[i]);
	}
	*p++ = 'A';
	*p++ = 'T';
	if(arg_count > BASE_CMD_LEN){
		for(i = 0 ; i < argc ; i++){
			s = argv[i];
			*p++ = '+';
			while(*s != 0){
				c = *s++;
				if(c == '"')
					caps ^= 1;
				if(caps && c >= 'a' && c <= 'z')
					c -= 0x20;
				*p++ = c;
			}
		}
	}
	*p++ = '\r';
	*p++ = '\n';
	*p++ = 0;
	result = serial_wri_dat(AT_PORTID, (const char *)aTxBuffer, arg_count);
	if(result < 0){
		syslog_1(LOG_ERROR, "AT command send error(%d) !", result);
	}
	printf("%d:%s", arg_count, aTxBuffer);
	return arg_count;
}

/*
 *  SW1
 */
void sw_int(void)
{
	syslog_0(LOG_NOTICE, "## sw_int() ##");
}

/*
 *  ᥤ󥿥
 */
void main_task(intptr_t exinf)
{
	T_SERIAL_RPOR k_rpor;
	queue_t *rxque;
	ER_UINT	ercd;
	int i, j, len;
	int dlen = 0;
	uint8_t ch;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	/*
	 * 塼Хåեν
	 */
	rxque = &rx_queue;
	rxque->size = RX_BUF_SIZE;
	rxque->clen  = 0;
	rxque->head = 0;
	rxque->tail = 0;
	rxque->pbuffer = aRxBuffer;

	/*
	 *  ESP-WROOM02ѥꥢݡȤν
	 */
	ercd = serial_opn_por(AT_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.(AT)",
									itron_strerror(ercd), SERCD(ercd));
		slp_tsk();
	}
	SVC_PERROR(serial_ctl_por(AT_PORTID,
							(IOCTL_FCSND | IOCTL_FCRCV)));
	while(dlen < 500000){
		serial_ref_por(AT_PORTID, &k_rpor);
		len = k_rpor.reacnt;
		if((rxque->size - rxque->clen) < len)
			len = rxque->size - rxque->clen;
		if(len > 0){
			if(rxque->head >= rxque->tail){
				if((rxque->size - rxque->head) < len)
					i = rxque->size - rxque->head;
				else
					i = len;
				j = serial_rea_dat(AT_PORTID, (char *)&rxque->pbuffer[rxque->head], i);
				rxque->head += j;
				if(rxque->head >= rxque->size)
					rxque->head -= rxque->size;
				rxque->clen += j;
				len -= j;
			}
			if(len > 0){
				j = serial_rea_dat(AT_PORTID, (char *)&rxque->pbuffer[rxque->head], len);
				rxque->head += j;
				if(rxque->head >= rxque->size)
					rxque->head -= rxque->size;
				rxque->clen += j;
				len -= j;
			}
		}
		while(rxque->clen > 0 && rx_mode != MODE_NOECHO){
			if(rx_mode == MODE_ECHO_CHAR){
				ch = rxque->pbuffer[rxque->tail];
				if(ch >= 0x7f)
					ch = '.';
				else if(ch < 0x20 && ch != '\r' && ch != '\n')
					ch = '.';
				putchar(ch);
			}
			else
				printf("%02x ", rxque->pbuffer[rxque->tail]);
			rxque->clen--;
			rxque->tail++;
			if(rxque->tail >= rxque->size)
				rxque->tail -= rxque->size;
			dlen++;
			if((dlen % 32) == 0 && rx_mode == MODE_ECHO_HEX){
				printf("\n");
				dly_tsk(50);
			}
		}
		dly_tsk(20);
	}
	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();

	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}
