/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: sample1.c 2416 2012-09-07 08:06:20Z ertl-hiro $
 */

/* 
 *  ץץ(1)
 *
 *  STM32F769I-DiscoveryΥǥХǥץࡥ
 *
 *  ץγ:
 *
 *  桼򲡤Ȥˡ˥塼Υǥץ¹Ԥ롥
 *  ǥץϡ2ѰդƤ롥ƥץ
 *  aspͥTOPPERS Base Platform(STM)V1.3.1ư롥
 *
 *  1.ǥաǥ(Ĥ)
 *  2.ޥϤȥǥ
 *
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <target_syssvc.h>
#include <string.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "sample1.h"

/*
 *  SILؿΥޥ
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

/* I2C TIMING Register define when I2C clock source is SYSCLK */
/* I2C TIMING is calculated from APB1 source clock = 50 MHz */
/* Due to the big MOFSET capacity for adapting the camera level the rising time is very large (>1us) */
/* 0x40912732 takes in account the big rising and aims a clock of 100khz */
#ifndef DISCOVERY_I2Cx_TIMING  
#define DISCOVERY_I2Cx_TIMING                      ((uint32_t)0x40912732)
#endif /* DISCOVERY_I2Cx_TIMING */

#define FMC_BANK2_BASE  ((uint32_t)(0x60000000 | 0x04000000))  

/* Private typedef -----------------------------------------------------------*/
/* Private define ------------------------------------------------------------*/

/* Private typedef -----------------------------------------------------------*/
typedef enum {
  TS_ACT_NONE = 0,
  TS_ACT_FREQ_DOWN,
  TS_ACT_FREQ_UP,
  TS_ACT_VOLUME_DOWN,
  TS_ACT_VOLUME_UP,
  TS_ACT_PAUSE = 0xFE
}TS_ActionTypeDef;

static const TS_PortControlBlock ts_pcb = {
	TADR_RCC_BASE+TOFF_RCC_AHB1ENR, RCC_AHB1ENR_GPIOHEN,	/* reset clk_base clk_bit */
	TADR_GPIOH_BASE, 9, GPIO_MODE_OUTPUT, GPIO_OTYPE_PP,	/*     base, pinno, mode, otype */
	GPIO_NOPULL, GPIO_SPEED_LOW, 1,							/*     pull, speed, release */
	TADR_RCC_BASE+TOFF_RCC_AHB1ENR, RCC_AHB1ENR_GPIOIEN,	/* int   clk_base, clk_bit */
	TADR_GPIOI_BASE, 9, GPIO_MODE_IT_FALLING,				/*     base, pinno, mode */
	GPIO_PULLUP, GPIO_SPEED_FAST							/*     pull speed */
};

volatile uint8_t key_in = 0;

LCD_Handler_t        LcdHandle;
TouchScreen_Handle_t TouchScreen_Handle;
STAUDIO_Handle_t     Audio_Handle;
LCD_DrawProp_t       DrawProp;

static void audio_error_callback(AUDIO_Handle_t *hsai)
{
	if(audio_status(hsai, AUDIO_OUT_BLOCK) > AUDIO_STATUS_READY){
		BSP_AUDIO_OUT_Error_CallBack();
	}
	if(audio_status(hsai, AUDIO_IN_BLOCK) > AUDIO_STATUS_READY){
		BSP_AUDIO_IN_Error_CallBack();
	}
}

/*
 *  桼ߥϥɥ
 */
void sw_handle(void)
{
	syslog_0(LOG_NOTICE, "## sw int ##");
	key_in++;
}

/*
 *  桼åå
 *  return  switch state (1 : active / 0 : Inactive)
 */
uint8_t CheckForUserInput(void)
{
	if(key_in > 0){
		key_in = 0;
	    dly_tsk(10);
    	return 1 ;
	}
	return 0;
}


/*
 *  ᥤ󥿥
 */
void main_task(intptr_t exinf)
{
	ER_UINT	ercd;
	int count = 0;
	I2C_Init_t     Init;
	I2C_Handle_t   *hi2c;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	/* Configure LCD */
	LcdHandle.fmc_base    = FMC_BANK2_BASE;
	LcdHandle.fmc_base    = FMC_BANK2_BASE;
	LcdHandle.rst_clk     = RCC_AHB1ENR_GPIOHEN;	/* RESET CLOCK BIT */
	LcdHandle.te_clk      = RCC_AHB1ENR_GPIOCEN;	/* LCD TE CLOCK BIT */
	LcdHandle.bl_clk      = RCC_AHB1ENR_GPIOHEN;	/* LCD BL CTRL CLOCK BIT */
	LcdHandle.rst_base    = TADR_GPIOH_BASE;		/* RESET PORT */
	LcdHandle.te_base     = TADR_GPIOC_BASE;		/* LCD TE PORT */
	LcdHandle.bl_base     = TADR_GPIOH_BASE;		/* LCD BL CTRL PORT */
	LcdHandle.gpio_speed  = GPIO_SPEED_FAST;		/* GPIO SPEED */
	LcdHandle.rst_pull    = GPIO_NOPULL;			/* RESET PULL TYPE */
	LcdHandle.bl_pull     = GPIO_PULLUP;			/* LCD BL CTRL PULL TYPE */
	LcdHandle.rst_pin     = 7;						/* RESET PIN */
	LcdHandle.te_pin      = 8;						/* LCD TE PIN */
	LcdHandle.bl_pin      = 11;						/* LCD BL CTRL PIN */
	LcdHandle.bl_active   = 1;

	BSP_LCD_Init();
	DrawProp.hlcd      = &LcdHandle;
	DrawProp.pFont     = &Font12;

	/* Init TS module */
	memset(&Init, 0, sizeof(I2C_Init_t));
	Init.Timing           = DISCOVERY_I2Cx_TIMING;
	Init.OwnAddress1      = 0;
	Init.AddressingMode   = I2C_ADDRESSINGMODE_7BIT;
	Init.DualAddressMode  = I2C_DUALADDRESS_DISABLE;
	Init.OwnAddress2      = 0;
	Init.GeneralCallMode  = I2C_GENERALCALL_DISABLE;
	Init.NoStretchMode    = I2C_NOSTRETCH_DISABLE;
    Init.semid   = I2CTSTRS_SEM;
    Init.semlock = I2CTSLOC_SEM;
	hi2c = i2c_init(I2C3_PORTID, &Init);
	if(hi2c == NULL){
		syslog_0(LOG_ERROR, "TS I2C INIT ERROR !");
		slp_tsk();
	}
	TouchScreen_Handle.hi2c      = hi2c;
	TouchScreen_Handle.pcb       = &ts_pcb;
	TouchScreen_Handle.ifunc     = NULL;
	TouchScreen_Handle.enable_it = 0;

	BSP_TS_Init(BSP_LCD_GetXSize(), BSP_LCD_GetYSize());
	BSP_LCD_LayerDefaultInit(0, 0);

	Touchscreen_Calibration(200, 170);

	/* Init Audio module */
    Init.semid   = I2CADTRS_SEM;
    Init.semlock = I2CADLOC_SEM;
	hi2c = i2c_init(I2C1_PORTID, &Init);
	if(hi2c == NULL){
		syslog_0(LOG_ERROR, "AUDIO I2C INIT ERROR !");
		slp_tsk();
	}

	Audio_Handle.hi2c          = hi2c;
	Audio_Handle.hsai          = NULL;
	Audio_Handle.saiid         = AUDIO2_PORTID;
	Audio_Handle.i2caddress    = AUDIO_I2C_ADDRESS;
	Audio_Handle.AudioInVolume = DEFAULT_AUDIO_IN_VOLUME;
	Audio_Handle.audio_fullout_func = BSP_AUDIO_OUT_TransferComplete_CallBack;
	Audio_Handle.audio_halfout_func = BSP_AUDIO_OUT_HalfTransfer_CallBack;
	Audio_Handle.audio_fullin_func  = BSP_AUDIO_IN_TransferComplete_CallBack;
	Audio_Handle.audio_halfin_func  = BSP_AUDIO_IN_HalfTransfer_CallBack;
	Audio_Handle.audio_error_func   = audio_error_callback;

	while(count < 20){

		AudioPlay_demo();

		AudioLoopback_demo();
		count++;
	}

	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();
	syslog(LOG_NOTICE, "Sample program ends.");
	SVC_PERROR(ext_ker());
	assert(0);
}
