/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2015-2016 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: audioutil.c 2416 2016-03-05 12:23:35Z roi $
 */
/*
 *  AUDIOѥ桼ƥƥ
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <kernel.h>
#include <t_stddef.h>
#include <t_syslog.h>
#include <target_syssvc.h>
#include "sdcard_play.h"
#include "mad.h"
#include "kernel_cfg.h"


#define AUDIO_I2C_ADDRESS                ((uint16_t)0x34)
#define AUDIO_COUNT_MAX                  0x7FFFFFFF

#define MPEG_BUFSZ           (64*1024)

static const uint16_t device_table[] = {
	OUTPUT_DEVICE_SPEAKER,
	OUTPUT_DEVICE_HEADPHONE,
	OUTPUT_DEVICE_BOTH,
	OUTPUT_DEVICE_AUTO,
	INPUT_DEVICE_DIGITAL_MICROPHONE_1,
	INPUT_DEVICE_DIGITAL_MICROPHONE_2,
	INPUT_DEVICE_INPUT_LINE_1,
	INPUT_DEVICE_INPUT_LINE_2
};

static const uint32_t option_table[] = {
	CODEC_PDWN_HW,
	CODEC_PDWN_SW
};

static MUSIC_INFO  Minfo;
static MUSIC_TRANS *pMTrans;
static STAUDIO_Handle_t  Audio_Handle;


static uint32_t
audio_setup_buffer(MUSIC_INFO *minfo, MUSIC_TRANS *mtrans, int mode)
{
	uint32_t  *s, *d;
	uint32_t  len, len2, i = 0;

	d = (uint32_t *)(minfo->buffer+(mode-SET_BUFFER0)*WAV_BUFFER_SIZE);
	s = (uint32_t *)&minfo->maddr[pMTrans->mtail];
	if(pMTrans->msize < WAV_BUFFER_SIZE)
		len  = pMTrans->msize;
	else
		len = WAV_BUFFER_SIZE;
	if(mtrans->transize >= mtrans->datasize){
		memset(d, 0, WAV_BUFFER_SIZE);
		return 0;
	}
	if((pMTrans->mtail + len) >= MUSIC_DATA_SIZE){
		len2 = MUSIC_DATA_SIZE - pMTrans->mtail;
		for(i = 0 ; i < len2 ; i += 4)
			*d++ = *s++;
		pMTrans->mtail = 0;
		s = (uint32_t *)minfo->maddr;
	}
	for(; i < len ; i += 4)
		*d++ = *s++;
	for(; i < WAV_BUFFER_SIZE; i += 4)
		*d++ = 0;
	pMTrans->mtail += len;
	pMTrans->msize -= len;
	return WAV_BUFFER_SIZE;
}

/*
 *  TRANSFAR END CALLBACK
 */
static void
audio_transfunc(AUDIO_Handle_t *haudio)
{
	if(pMTrans == NULL)
		return;
	pMTrans->transize += WAV_BUFFER_SIZE;
	audio_setup_buffer(pMTrans->minfo, pMTrans, SET_BUFFER1);
	iset_flg(AUDIO_FLG, SET_BUFFER1);
}

static void
audio_transhfunc(AUDIO_Handle_t *haudio)
{
	if(pMTrans == NULL)
		return;
	pMTrans->transize += WAV_BUFFER_SIZE;
	audio_setup_buffer(pMTrans->minfo, pMTrans, SET_BUFFER0);
	iset_flg(AUDIO_FLG, SET_BUFFER0);
}

void
audio_handle(AUDIO_Handle_t *haudio)
{
	syslog_1(LOG_ERROR, "## audio int[%08x] ##", haudio);
}

/*
 *  AUDIO
 */
MUSIC_INFO *
AUDIO_Hard_Init(ID id, I2C_Handle_t *hi2c)
{
	Audio_Handle.hsai          = NULL;
	Audio_Handle.hi2c          = hi2c;
	Audio_Handle.saiid         = id;
	Audio_Handle.i2caddress    = AUDIO_I2C_ADDRESS;
	Audio_Handle.AudioInVolume = DEFAULT_AUDIO_IN_VOLUME;
	Audio_Handle.audio_fullout_func = NULL;
	Audio_Handle.audio_halfout_func = NULL;
	Audio_Handle.audio_fullin_func  = NULL;
	Audio_Handle.audio_halfin_func  = NULL;
	Audio_Handle.audio_error_func   = audio_handle;
	Minfo.haudio = &Audio_Handle;
	Minfo.maddr  = (uint8_t *)malloc(MUSIC_DATA_SIZE);
	Minfo.buffer = (uint8_t *)malloc(WAV_BUFFER_SIZE*2);
	AUDIO_OUT_Init(Minfo.haudio, AUDIO_OUT_DEVICE_HEADPHONE, 50, 22050, 32);
	memset(Minfo.buffer, 0, 0x4000);
	AUDIO_OUT_Play(Minfo.haudio, (uint16_t*)Minfo.buffer, 0x4000, 0);
	dly_tsk(1000);
	AUDIO_OUT_Stop(Minfo.haudio, POWERDOWN_HW);
	return &Minfo;
}

/*
 *  AUDIO OUT
 */
ER
AUDIO_OUT_Init(STAUDIO_Handle_t *haudio, AudioDevice Dev, uint8_t Volume, uint32_t AudioFreq, uint16_t framesize)
{
	syslog_3(LOG_NOTICE, "AUDIO_OUT_Init vol(%d) freq(%d) size(%d) !", Volume, AudioFreq, framesize);
	if(haudio == NULL)
		return E_PAR;
	return staudio_out_init(haudio, device_table[Dev], Volume, AudioFreq, framesize);
}

/*
 *  AUDIO OUT PLAY
 */
ER
AUDIO_OUT_Play(STAUDIO_Handle_t *haudio, uint16_t* pBuffer, uint32_t Size, uint32_t mode)
{
	if(haudio == NULL)
		return E_PAR;
	haudio->audio_fullout_func = audio_transfunc;
	if(mode == 1)
		haudio->audio_halfout_func = audio_transhfunc;
	else
		haudio->audio_halfout_func = NULL;
	return staudio_out_play(haudio, pBuffer, Size);
}

/*
 *  AUDIO OUT STOP
 */
ER
AUDIO_OUT_Stop(STAUDIO_Handle_t *haudio, CodecPowerDown Option)
{
	if(haudio == NULL)
		return E_PAR;
	return staudio_out_stop(haudio, option_table[Option]);
}


/*
 *  WAVǡž
 */
ER
wav_transfar(MUSIC_TRANS *mtrans)
{
	static Riff_Header    rhead;
	static Data_Header    dhead;
	MUSIC_INFO *minfo = mtrans->minfo;
	FLGPTN   flgptn;
	ER       ercd = E_OK;
	int      len, no;

	pMTrans = mtrans;
	mtrans->mcount = 0;
	mtrans->mhead  = 0;
	mtrans->mtail  = 0;
	mtrans->transize = 0;
	no = fread(&rhead, sizeof(Riff_Header), 1, minfo->fileid);
	syslog_3(LOG_NOTICE, "## no(%d)[%08x][%08x] ##", no, rhead.chank_size, &rhead);
	if(rhead.chank_size > 16){
		fread(minfo->maddr, rhead.chank_size-16, 1, minfo->fileid);
	}
	no = fread(&dhead, sizeof(Data_Header), 1, minfo->fileid);
	while(dhead.id[0] != 'd' || dhead.id[1] != 'a' || dhead.id[2] != 't' || dhead.id[3] != 'a'){
		no = fread(minfo->maddr, dhead.data_size, 1, minfo->fileid);
		no = fread(&dhead, sizeof(Data_Header), 1, minfo->fileid);
		break;
	}
	syslog_4(LOG_NOTICE, "## file size[%08x] channels(%d) audiofreq(%d) bitswidth(%d) ##", rhead.file_size, rhead.channels, rhead.audiofreq, rhead.bitswidth);
	syslog_2(LOG_NOTICE, "## blocksize(%d) bytepersec(%d)  ##", rhead.blocksize, rhead.bytepersec);
	syslog_1(LOG_NOTICE, "## size[%08x] ##", dhead.data_size);
	minfo->status |= AUDIO_ACTIVE;
	mtrans->datasize   = dhead.data_size;
	mtrans->channels   = rhead.blocksize / 2;
	mtrans->samplerate = rhead.audiofreq / mtrans->channels;
	mtrans->func(mtrans);
	mtrans->mcount++;

	if(mtrans->datasize <= MUSIC_DATA_SIZE)
		minfo->transize = mtrans->datasize;
	else
		minfo->transize = MUSIC_DATA_SIZE;
	if(mtrans->datasize <= (WAV_BUFFER_SIZE*2)){
		minfo->buffsize = minfo->datasize;
		minfo->playmode = 0;
	}
	else{
		minfo->buffsize = (WAV_BUFFER_SIZE*2);
		minfo->playmode = 1;
	}
	no = fread(minfo->maddr, 1, minfo->transize, minfo->fileid);
	mtrans->mhead = no;
	mtrans->msize = no;
	audio_setup_buffer(minfo, mtrans, SET_BUFFER0);
	audio_setup_buffer(minfo, mtrans, SET_BUFFER1);
	mtrans->func(mtrans);
	mtrans->mcount++;

	while(mtrans->transize < mtrans->datasize){
		ercd = twai_flg(AUDIO_FLG, SET_BUFFER, TWF_ORW, &flgptn, 5000);
		if(ercd != E_OK){
			if((minfo->status & AUDIO_PHASE) != 0)
				continue;
			syslog_1(LOG_ERROR, "audio timeout(%d) !", ercd);
			break;
		}
		len = MUSIC_DATA_SIZE - mtrans->msize;
		no = minfo->datasize - minfo->transize;
		if(no <= 0)
			continue;
		if(len > no)
			len = no;
		if((mtrans->mhead + len) >= MUSIC_DATA_SIZE){
			no = MUSIC_DATA_SIZE - mtrans->mhead;
			if(no > 0)
				no = fread(&minfo->maddr[mtrans->mhead], 1, no, minfo->fileid);
			mtrans->mhead += no - MUSIC_DATA_SIZE;
			mtrans->msize += no;
			minfo->transize += no;
			len -= no;
		}
		no = fread(&minfo->maddr[mtrans->mhead], 1, len, minfo->fileid);
		mtrans->mhead += no;
		mtrans->msize += no;
		minfo->transize += len;
		mtrans->func(mtrans);
		mtrans->mcount++;
	}
	pMTrans = NULL;
	minfo->status = 0;
	return ercd;
}

/*
 * This is the input callback. The purpose of this callback is to (re)fill
 * the stream buffer which is to be decoded. In this example, an entire file
 * has been mapped into memory, so we just call mad_stream_buffer() with the
 * address and length of the mapping. When this callback is called a second
 * time, we are finished decoding.
 */
static enum
mad_flow input(void *data,
		    struct mad_stream *stream)
{
	MUSIC_TRANS *mtrans = data;
	MUSIC_INFO  *minfo;
	int len, len2;

	minfo = mtrans->minfo;
	if(minfo->transize >= minfo->datasize || (minfo->status & AUDIO_STOP) != 0)
		return MAD_FLOW_STOP;

	if(stream->next_frame) {
		BCOPY(stream->next_frame, minfo->start,
			minfo->length = &minfo->start[minfo->length] - stream->next_frame);
	}

	len = fread(minfo->start + minfo->length, 1, MPEG_BUFSZ - minfo->length, minfo->fileid);
	syslog_3(LOG_NOTICE, "input:read len[%08x] mtrans->length[%08x][%08x]!", len, minfo->length, minfo->transize);

	if(len == 0){
		if((MPEG_BUFSZ - minfo->length) < MAD_BUFFER_GUARD)
			syslog_2(LOG_ERROR, "input:buffer->length[%08x] GUARD[%08x] !", minfo->length, MAD_BUFFER_GUARD);

		len2 = len;
		while (len2 < MAD_BUFFER_GUARD && (minfo->length+len) < MPEG_BUFSZ)
			minfo->start[minfo->length + len2++] = 0;
	}

	mad_stream_buffer(stream, minfo->start, minfo->length += len);
	minfo->transize += len;
	if(len == 0 && minfo->transize < (minfo->datasize-1024)){
		syslog_2(LOG_ERROR, "input:read error minfo->transize[%08x][%08x] !", minfo->transize, minfo->datasize);
		return MAD_FLOW_STOP;
	}
	else
		return MAD_FLOW_CONTINUE;
}

/*
 * The following utility routine performs simple rounding, clipping, and
 * scaling of MAD's high-resolution samples down to 16 bits. It does not
 * perform any dithering or noise shaping, which would be recommended to
 * obtain any exceptional audio quality. It is therefore not recommended to
 * use this routine if high-quality output is desired.
 */
static inline
signed int scale(mad_fixed_t sample)
{
	/* round */
	sample += (1L << (MAD_F_FRACBITS - 16));

	/* clip */
	if (sample >= MAD_F_ONE)
		sample = MAD_F_ONE - 1;
	else if (sample < -MAD_F_ONE)
		sample = -MAD_F_ONE;

	/* quantize */
	return sample >> (MAD_F_FRACBITS + 1 - 16);
}

/*
 * This is the output callback function. It is called after each frame of
 * MPEG audio data has been completely decoded. The purpose of this callback
 * is to output (or play) the decoded PCM audio.
 */
static enum
mad_flow output(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
	MUSIC_TRANS *mtrans = data;
	MUSIC_INFO  *minfo;
	FLGPTN   flgptn;
	ER       ercd = E_OK;
	unsigned int nsamples;
	mad_fixed_t const *left_ch, *right_ch;

	minfo = mtrans->minfo;
	if(mtrans->mcount == 0){
		mtrans->channels   = pcm->channels;
		mtrans->samplerate = pcm->samplerate / mtrans->channels;
		mtrans->func(mtrans);
		mtrans->mcount++;
	}
	if((minfo->status & AUDIO_STOP) != 0)
		return MAD_FLOW_CONTINUE;

	/* pcm->samplerate contains the sampling frequency */

	nsamples  = pcm->length;
	left_ch   = pcm->samples[0];
	right_ch  = pcm->samples[1];

	while (nsamples--) {
		signed int sample;

		/* output sample(s) in 16-bit signed little-endian PCM */

		while((volatile int)mtrans->msize >= (MUSIC_DATA_SIZE-3)){
			if(mtrans->mcount == 1){
				mtrans->func(mtrans);
				mtrans->mcount++;
			}
			ercd = twai_flg(AUDIO_FLG, SET_BUFFER, TWF_ORW, &flgptn, 5000);
			if(ercd != E_OK){
				if((minfo->status & AUDIO_PHASE) != 0)
					continue;
				syslog_1(LOG_ERROR, "output timeout(%d) !", ercd);
				minfo->transize  = minfo->datasize;
				minfo->status   |= AUDIO_STOP;
				mtrans->transize = AUDIO_COUNT_MAX;
				return MAD_FLOW_CONTINUE;
			}
		}
		sample = scale(*left_ch++);
		loc_cpu();
		minfo->maddr[mtrans->mhead++] = (sample >> 0) & 0xff;
		minfo->maddr[mtrans->mhead++] = (sample >> 8) & 0xff;
		if(mtrans->mhead >= MUSIC_DATA_SIZE)
			mtrans->mhead -= MUSIC_DATA_SIZE;
		mtrans->msize += 2;
		mtrans->datasize += 2;
		if (mtrans->channels == 2) {
			sample = scale(*right_ch++);
			minfo->maddr[mtrans->mhead++] = (sample >> 0) & 0xff;
			minfo->maddr[mtrans->mhead++] = (sample >> 8) & 0xff;
			if(mtrans->mhead >= MUSIC_DATA_SIZE)
				mtrans->mhead -= MUSIC_DATA_SIZE;
			mtrans->msize += 2;
			mtrans->datasize += 2;
		}
		unl_cpu();
	}

	return MAD_FLOW_CONTINUE;
}

/*
 * This is the error callback function. It is called whenever a decoding
 * error occurs. The error is indicated by stream->error; the list of
 * possible MAD_ERROR_* errors can be found in the mad.h (or stream.h)
 * header file.
 */
static enum
mad_flow error(void *data,
		    struct mad_stream *stream,
		    struct mad_frame *frame)
{
	MUSIC_TRANS *mtrans = data;
	MUSIC_INFO  *minfo;

	minfo = mtrans->minfo;
	syslog_3(LOG_ERROR, "decoding error 0x%04x (%s) at byte offset %u",
		stream->error, mad_stream_errorstr(stream),
		stream->this_frame - minfo->start);

	/* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

	return MAD_FLOW_CONTINUE;
}

/*
 *  MP3 DECODE
 */
ER
mp3_decode(MUSIC_TRANS *mtrans)
{
	MUSIC_INFO  *minfo;
	struct mad_decoder decoder;
	FLGPTN   flgptn;
	int result;

	/* initialize our private message structure */
	pMTrans = mtrans;
	minfo = mtrans->minfo;
	minfo->start  = malloc(MPEG_BUFSZ);
	minfo->length = 0;
	minfo->buffsize = (WAV_BUFFER_SIZE*2);
	minfo->playmode = 1;
	minfo->transize = 0;
	syslog_1(LOG_NOTICE, "## fsize[%08x] ##", minfo->datasize);

	mtrans->mcount = 0;
	mtrans->mhead  = 0;
	mtrans->mtail  = 0;
	mtrans->msize  = 0;
	mtrans->datasize = 0;
	mtrans->transize = 0;
	minfo->status |= AUDIO_ACTIVE;

	/* configure input, output, and error functions */

	mad_decoder_init(&decoder, mtrans,
		   input, 0 /* header */, 0 /* filter */, output,
		   error, 0 /* message */);

	/* start decoding */

	result = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

	/* release the decoder */

	mad_decoder_finish(&decoder);
	syslog_2(LOG_NOTICE, "## data[%08x] trans[%08x] ##", minfo->datasize, minfo->transize);
	while(mtrans->transize < mtrans->datasize){
		twai_flg(AUDIO_FLG, SET_BUFFER, TWF_ORW, &flgptn, 1000);
	}
	free(minfo->start);
	pMTrans = NULL;
	minfo->status = 0;

	if(result == 0)
		return E_OK;
	else
		return E_OBJ;
}

