/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2016 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: storagedevice.c,v 1.3 2016/04/24 11:50:55 roi Exp $
 */

#include <stdlib.h>
#include <string.h>
#include "kernel_cfg.h"
#include "storagedevice.h"

#define SENSE_DEVICE    (SDEV_ACTIVE|SDEV_INSERTCHK)

/*
 *  Storage Device Mnanager
 */
StorageDeviceHead_t SDeviceHead;

/*
 *  ǥХñ̤ι¤ΰ
 */
static StorageDevice_t SDevice[NUM_STORAGEDEVICE];



void sdev_init(intptr_t exinf)
{
	unsigned int no;

	for(no = 0 ; no < NUM_STORAGEDEVICE ; no++){
		memset(&SDevice[no], 0, sizeof(StorageDevice_t));
	}
	memset(&SDeviceHead, 0, sizeof(StorageDeviceHead_t));
	SDeviceHead._psd            = &SDevice[0];
	SDeviceHead._default_device = DEFAULT_DEVNO;
    SDeviceHead._sdev_active    = true;
}

void sdev_terminate(void)
{
	SDeviceHead._sdev_active    = false;
}

ER SDMSetupDevice(int16_t devno, StorageDevice_t **ppsdev)
{
	StorageDevice_t *psdev;
	unsigned int    no;

	if(devno < MINIMUM_DEVNO || devno > MAXIMUM_DEVNO)
		return E_ID;
	if(SDeviceHead._num_activedev >= NUM_STORAGEDEVICE)
		return E_OBJ;
	for(no = 0, psdev = SDeviceHead._psd ; no < SDeviceHead._num_activedev ; no++, psdev++){
		if(devno == psdev->_sdev_devno)
			return E_ID;
	}
	psdev->_sdev_attribute = SDEV_ACTIVE;
	psdev->_sdev_devno     = devno;
	*ppsdev = psdev;
	SDeviceHead._num_activedev++;
	return E_OK;
}

ER_ID SDMGetDeviceNo(const char **ppathname)
{
	ER_ID devno = DEFAULT_DEVNO;
	const char *p = *ppathname;

	if(*p >= '0'&& *p <= '9'){
		devno = *p++ - '0';
		while(*p >= '0' && *p <= '9'){
			devno *= 10;
			devno  = *p++ - '0';
		}
		if(devno < MINIMUM_DEVNO || devno > MAXIMUM_DEVNO)
			devno = E_ID;
	}
	if(*p == ':')
		p++;
	*ppathname = p;
	return devno;
}

StorageDevice_t *SDMGetStorageDevice(int devno)
{
	StorageDevice_t *psdev = SDeviceHead._psd;
	unsigned int no;

    for(no = 0 ; no < SDeviceHead._num_activedev ; no++){
		if((psdev->_sdev_attribute & SDEV_ACTIVE) != 0
			 && psdev->_sdev_devno == devno)
			return psdev;
		psdev++;
	}
	return NULL;
}

ER SDMEmploy(StorageDevice_t *psdev, bool_t sw)
{
	if(psdev == 0)
		return E_PAR;
	if((psdev->_sdev_attribute & SDEV_ACTIVE) == 0)
		return E_PAR;
	if(sw)
		psdev->_sdev_attribute |= SDEV_EMPLOY;
	else
		psdev->_sdev_attribute &= ~SDEV_EMPLOY;
	return E_OK;
}

void SDMSence_task(intptr_t exinf)
{
	StorageDevice_t *psdev;
	unsigned int dno;
	bool_t       employ;

	while(SDeviceHead._sdev_active == true){
		psdev = SDeviceHead._psd;
		for(dno = 0 ; dno < SDeviceHead._num_activedev ; dno++, psdev++){
			if((psdev->_sdev_attribute & SENSE_DEVICE) == SENSE_DEVICE){
				employ = (psdev->_sdev_attribute & SDEV_EMPLOY) != 0;
				if(!employ && psdev->pdevf->_sdevf_sense(psdev, employ) != 0){
					psdev->_sdev_attribute |= SDEV_EMPLOY;
                    if(psdev->_sdev_notice)
						psdev->_sdev_notice(psdev, true);
				}
				else if((psdev->_sdev_attribute & SDEV_CHKREMOVE) != 0 && employ){
					if(psdev->pdevf->_sdevf_sense(psdev, employ) != 0){
						psdev->_sdev_attribute &= ~SDEV_EMPLOY;
						psdev->_sdev_attribute |= SDEV_ONEEXIT;
						psdev->_sdev_instimer   = 0;
	                    if(psdev->_sdev_notice)
							psdev->_sdev_notice(psdev, false);
					}
				}
			}
		}
		dly_tsk(SENSE_TIME);
#ifdef SDEV_SENSE_ONETIME
		break;
#endif
	}
}


