/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 *
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2010 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2012-2015 by GT Development Center RICOH COMPANY,LTD. JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: clock.c,v 1.4 2019/01/27 21:54:41 roi Exp $
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <target_syssvc.h>
#define _SYS__STDINT_H
#include <time.h>
#include "kernel_cfg.h"
#include "device.h"
#include "rtc.h"
#include "monitor.h"

#if defined(RTC_CLOCK_LSE)
#define RTC_CLOCK   32768	/* LSE_VALUE */
#else
#define RTC_CLOCK   32000	/* LSI_VALUE */
#endif

static int_t date_func(int argc, char **argv);
static int_t time_func(int argc, char **argv);
static int_t wakeup_func(int argc, char **argv);
static int_t clock_func(int argc, char **argv);

/*
 *  ңԣåޥɥơ֥
 */
static const COMMAND_INFO rtc_command_info[] = {
	{"DATE",	date_func    },	/* ˤ */
	{"TIME",	time_func    },	/*  */
	{"WAKEUP",	wakeup_func  },	/* WAKEUP */
	{"CLOCK",	clock_func   }	/*  */
};

#define NUM_RTC_CMD     (sizeof(rtc_command_info)/sizeof(COMMAND_INFO))

static const char rtc_name[] = "RTC";
static const char rtc_help[] =
"  Rtc     DATE  (year #) (month #) (day #)\n"
"          TIME  (hour #) (min   #) (sec #)\n"
#ifdef USE_RTC_WAKEUP_CMD
"          WAKEUP time(ms)\n"
#endif
"          CLOCK DATE:TIME\n";

static const char monthday[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

static COMMAND_LINK rtc_command_link = {
	NULL,
	NUM_RTC_CMD,
	rtc_name,
	NULL,
	rtc_help,
	&rtc_command_info[0]
};

static RTC_Wakeup_t rtc_wup;

/*
 *  atox
 */
static int atoix(char *s)
{
	int c, val = 0;

	while((c = *s++) != 0){
		if(c >= '0' && c <= '9')
			val = val * 10 + (c - '0');
	}
	return val;
}

#ifdef USE_RTC_WAKEUP_CMD
/*
 *  WAKUP EVENT
 */
static void wakeup_event(void)
{
	syslog_0(LOG_NOTICE, "rtc wakeup event !");
}
#endif

/*
 *  mktimeؿ
 */
time_t mktime(struct tm *ptm)
{
	int day = ptm->tm_year *365;
	int yday, i; 

	day += (ptm->tm_year+2) / 4;
	for(i = 1, yday = 0 ; i < ptm->tm_mon ; i++)
		yday += monthday[i-1];
	if(ptm->tm_mon > 2 && ((ptm->tm_year+2) % 4) == 0)
		yday++;
	yday += ptm->tm_mday - 1;
	day  += yday;
	ptm->tm_wday = (day+4) % 7;
	ptm->tm_yday = yday+1;
	syslog_3(LOG_INFO, "yday[%d] wday[%d] day[%d]", yday+1, ptm->tm_wday, ptm->tm_yday);
	return day * (24*60*60) + ptm->tm_hour * 60*60 + ptm->tm_min * 60 + ptm->tm_sec;
}

/*
 * gmtime_rؿ
 */
struct tm *gmtime_r(const time_t *pt, struct tm *ptm)
{
	int day = *pt / (24*60*60);
	int time, i, yday, mday;

	time = *pt - day * 24 * 60 * 60;
	ptm->tm_wday = (day+4) % 7;
	for(i = 365, ptm->tm_year = 0 ; i < day ; i += 365){
		if(((ptm->tm_year+2) % 4) == 0)
			i++;
		ptm->tm_year++;
	}

	ptm->tm_yday = yday = day - i + 365 + 1;
	for(ptm->tm_mon = 1 ; ptm->tm_mon <= 12 && yday > 0 ; ptm->tm_mon++){
		mday = monthday[ptm->tm_mon-1];
		if(ptm->tm_mon == 2 && ((ptm->tm_year+2) % 4) == 0)
			mday++;
		if(yday < mday){
			ptm->tm_mday = yday;
			break;
		}
		yday -= mday;
	}
	ptm->tm_hour = time / (60*60);
	time -= ptm->tm_hour * 60 * 60;
	ptm->tm_min  = time / 60;
	ptm->tm_sec = time % 60;
	return ptm;
}

/*
 *  RTCޥؿ
 */
void rtc_info_init(intptr_t exinf)
{
#ifdef USE_RTC_WAKEUP_CMD
	rtc_wakeup_init(&rtc_wup, wakeup_event);
#endif
	setup_command(&rtc_command_link);
}

/*
 *  ꥳޥɴؿ
 */
static int_t date_func(int argc, char **argv)
{
	struct tm timedate;
	int arg1 = 2016;

	rtc_get_time((struct tm2 *)&timedate);
	if(argc > 1)
		arg1 = atoix(argv[1]);
	if(argc > 2)
		timedate.tm_mon = atoix(argv[2]);
	else
		timedate.tm_mon = 1;
	if(argc > 3)
		timedate.tm_mday = atoix(argv[3]);
	else
		timedate.tm_mday = 1;
	printf("%04d/%02d/%02d\n", arg1, timedate.tm_mon, timedate.tm_mday);
	timedate.tm_year = arg1-1970;
	timedate.tm_isdst = 0;
	mktime((struct tm *)&timedate);
	rtc_set_time((struct tm2 *)&timedate);
	return 0;
}

/*
 *  ꥳޥɴؿ
 */
static int_t time_func(int argc, char **argv)
{
	struct tm2 timedate;

	rtc_get_time(&timedate);
	if(argc > 1)
		timedate.tm_hour = atoix(argv[1]);
	else
		timedate.tm_hour = 0;
	if(argc > 2)
		timedate.tm_min = atoix(argv[2]);
	else
		timedate.tm_min = 0;
	if(argc > 3)
		timedate.tm_sec = atoix(argv[3]);
	else
		timedate.tm_sec = 0;
	printf("%02d:%02d:%02d\n", timedate.tm_hour, timedate.tm_min, timedate.tm_sec);
	timedate.tm_isdst = 0;
	mktime((struct tm *)&timedate);
	rtc_set_time(&timedate);
	return 0;
}

/*
 *  WAKEUP٥ؿ
 */
static int_t wakeup_func(int argc, char **argv)
{
	uint32_t count = 0;
	if(argc > 1){
		count = atoix(argv[1]);
		count = (count * (RTC_CLOCK >> rtc_wup.wakeuptimerPrescaler)) / 1000;
		if(count <= 0xFFFF)
			rtc_setup_wakeuptime(count);
		else
			printf("wakeup renge error(%d) !", count);
		return 0;
	}
	else
		return 1;
}

/*
 *  Фޥɴؿ
 */
static int_t clock_func(int argc, char **argv)
{
	struct tm2 timedate;

	rtc_get_time(&timedate);
	mktime(&timedate);
	printf("        %04d/%02d/%02d\n", timedate.tm_year+1970, timedate.tm_mon, timedate.tm_mday);
	printf("y[%d][%d] %02d:%02d:%02d\n", timedate.tm_yday, timedate.tm_wday, timedate.tm_hour, timedate.tm_min, timedate.tm_sec);
	return 0;
}

