/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2018 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: device.c 698 2018-11-06 20:54:48Z roi $
 */
/*
 * STM32L4xxѥǥХɥ饤
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <target_syssvc.h>
#include "kernel_cfg.h"
#include "device.h"


/*
 *  SILؿΥޥ
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#if !defined(HSE_VALUE)
#define HSE_VALUE               8000000		/* Value of the External oscillator in Hz */
#endif /* HSE_VALUE */
#if !defined(HSI_VALUE)
#define HSI_VALUE               16000000	/* Value of the Internal oscillator in Hz*/
#endif /* HSI_VALUE */

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

uint_t dipsw_value;
void (*sw_func)(int arg);
void (*exti5_func[NUM_EXTI5_FUNC])(void) = { 0 };
void (*exti10_func[NUM_EXTI10_FUNC])(void) = { 0 };


/*
 *  GPIO⡼ɤ
 */
#define GPIO_MODE               0x00000003
#define EXTI_MODE               0x10000000
#define GPIO_MODE_IT            0x00010000
#define GPIO_MODE_EVT           0x00020000
#define RISING_EDGE             0x00100000
#define FALLING_EDGE            0x00200000
#define GPIO_OUTPUT_TYPE        0x00000010

static const uint32_t gpio_index[] = {
	TADR_GPIOA_BASE,			/* index 0 */
	TADR_GPIOB_BASE,			/* index 1 */
	TADR_GPIOC_BASE,			/* index 2 */
	TADR_GPIOD_BASE,			/* index 3 */
	TADR_GPIOE_BASE,			/* index 4 */
	TADR_GPIOF_BASE,			/* index 5 */
	TADR_GPIOG_BASE,			/* index 6 */
	TADR_GPIOH_BASE				/* index 7 */
};

#define NUM_OF_GPIOPORT (sizeof(gpio_index)/sizeof(uint32_t))

/*
 *  GPIOνؿ
 */
void
gpio_setup(uint32_t base, GPIO_Init_t *init, uint32_t pin)
{
	uint32_t iocurrent = 1<<pin;
	uint32_t temp = 0x00;
	uint32_t index;

	for(index = 0 ; index < NUM_OF_GPIOPORT ; index++){
		if(gpio_index[index] == base)
			break;
	}
	if(index == NUM_OF_GPIOPORT)
		return;

	/*
	 *  GPIO⡼
	 */
	/* 륿͡ȡե󥯥󡦥⡼ */
	temp = sil_rew_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4));
	temp &= ~((uint32_t)0xF << ((uint32_t)(pin & (uint32_t)0x07) * 4)) ;
	if(init->mode == GPIO_MODE_AF)
		temp |= ((uint32_t)(init->alternate) << (((uint32_t)pin & (uint32_t)0x07) * 4));
	sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4), temp);

	/*  ϥ⡼ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_MODER), (GPIO_MODER_MODER2 << (pin * 2)), ((init->mode & GPIO_MODE) << (pin * 2)));

	/*  ϥ⡼ */
	if(init->mode == GPIO_MODE_OUTPUT || init->mode == GPIO_MODE_AF){
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OSPEEDR), (GPIO_OSPEEDER_OSPEEDR2 << (pin * 2)), (init->speed << (pin * 2)));
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OTYPER), (GPIO_OTYPER_OT << pin), (init->otype << pin));
	}

	/*  ץ륢åסץ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_PUPDR), (GPIO_PUPDR_PUPDR2 << (pin * 2)), (init->pull << (pin * 2)));

#ifdef TOFF_GPIO_ASCR
	/* ʥ ADC */
	if(init->mode == GPIO_MODE_ANALOG)
		sil_andw_mem((uint32_t *)(base+TOFF_GPIO_ASCR), (1<<pin));
	else if(init->mode == GPIO_MODE_ANALOG_AD)
		sil_orw_mem((uint32_t *)(base+TOFF_GPIO_ASCR), (1<<pin));
#endif

	/*
	 *  EXTI⡼
	 */
	if((init->mode & EXTI_MODE) == EXTI_MODE){
		/*  SYSCFGå */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB2ENR), RCC_APB2ENR_SYSCFGEN);
		temp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB2ENR));

		temp = sil_rew_mem((uint32_t *)(TADR_SYSCFG_BASE+TOFF_SYSCFG_EXTICR+(pin & 0x0C)));
		temp &= ~(0x0F << (4 * (pin & 0x03)));
		temp |= (index << (4 * (pin & 0x03)));
		sil_wrw_mem((uint32_t *)(TADR_SYSCFG_BASE+TOFF_SYSCFG_EXTICR+(pin & 0x0C)), temp);

		if((init->mode & GPIO_MODE_IT) == GPIO_MODE_IT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), iocurrent);

		if((init->mode & GPIO_MODE_EVT) == GPIO_MODE_EVT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EMR1), iocurrent);

		if((init->mode & RISING_EDGE) == RISING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);

		if((init->mode & FALLING_EDGE) == FALLING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
	}
}


/*
 *  ƥ९å֤
 *  return SYSCLK(Hz)
 */
uint32_t
get_sysclock(uint8_t *sws)
{
	static const uint32_t MSIRangeTable[12] = {100000, 200000, 400000, 800000, 1000000, 2000000, \
                                      4000000, 8000000, 16000000, 24000000, 32000000, 48000000};
	uint32_t cfgr = 0, pllm = 0, pllvco = 0;
	uint32_t msirange = 0, pllsource = 0, pllr = 2;
	uint32_t sysclockfreq = 0;

	cfgr = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR));

	/*
	 *  SYSCLKФ
	 */
	switch (cfgr & RCC_CFGR_SWS){
    case RCC_CFGR_SWS_HSI:	/* HSI used as system clock source */
		sysclockfreq =  HSI_VALUE;
		break;
	case RCC_CFGR_SWS_HSE:	/* HSE used as system clock */
		sysclockfreq = HSE_VALUE;
		break;
	default:
		if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_MSIRGSEL) != 0)
			msirange = (sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_MSIRANGE) >> 4;
		else
			msirange = (sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CSR)) & RCC_CSR_MSISRANGE) >> 8;
		/*MSI frequency range in HZ*/
		msirange = MSIRangeTable[msirange];

		if((cfgr & RCC_CFGR_SWS) == RCC_CFGR_SWS_MSI)
			sysclockfreq = msirange;
		else{
			/*
			 *  PLLƥ९åξ
			 */
			pllsource = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR)) & RCC_PLLCFGR_PLLSRC;
			pllm = ((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR)) & RCC_PLLCFGR_PLLM) >> 4) + 1;

			switch (pllsource){
			case RCC_PLLCFGR_PLLSRC_HSI:  /* HSI used as PLL clock source */
				pllvco = (HSI_VALUE / pllm) * ((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR)) & RCC_PLLCFGR_PLLN) >> 8);
				break;
			case RCC_PLLCFGR_PLLSRC_HSE:  /* HSE used as PLL clock source */
				pllvco = (HSE_VALUE / pllm) * ((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR)) & RCC_PLLCFGR_PLLN) >> 8);
				break;
			case RCC_PLLCFGR_PLLSRC_MSI:  /* MSI used as PLL clock source */
			default:
				pllvco = (msirange / pllm) * ((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR)) & RCC_PLLCFGR_PLLN) >> 8);
				break;
			}
			pllr = (((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR)) & RCC_PLLCFGR_PLLR) >> 25) + 1) * 2;
			sysclockfreq = pllvco/pllr;
		}
	}
	if(sws != NULL)
		*sws = (cfgr & RCC_CFGR_SWS) >> 2;
	return sysclockfreq;
}


/*
 *  DMACؿ
 */
#define NUM_CHDMA       14

#define DMA_TRS_TIMEOUT 2000	/* 2 */

#define DMAmux_Channel(h)   (TADR_DMAMUX1_BASE+((h)->chid2)*4)
#define DMAmux_RreqGEN(h)   (TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GEN+((h)->genreqno)*4)

static const uint32_t dma_base[NUM_CHDMA] = {
	TADR_DMA1_CH1_BASE,
	TADR_DMA1_CH2_BASE,
	TADR_DMA1_CH3_BASE,
	TADR_DMA1_CH4_BASE,
	TADR_DMA1_CH5_BASE,
	TADR_DMA1_CH6_BASE,
	TADR_DMA1_CH7_BASE,
	TADR_DMA2_CH1_BASE,
	TADR_DMA2_CH2_BASE,
	TADR_DMA2_CH3_BASE,
	TADR_DMA2_CH4_BASE,
	TADR_DMA2_CH5_BASE,
	TADR_DMA2_CH6_BASE,
	TADR_DMA2_CH7_BASE
};

static DMA_Handle_t *pDmaHandler[NUM_CHDMA];

/*
 *  CHANNEL DMAؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_init(DMA_Handle_t *hdma)
{
	uint32_t tmp = 0;
	uint32_t i, index;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < NUM_CHDMA ; i++){
		if(dma_base[i] == hdma->cbase)
			break;
	}
	if(i == NUM_CHDMA)
		return E_PAR;
	hdma->chid  = i % 7;
	hdma->chid2 = i;
	pDmaHandler[i] = hdma;
	hdma->base = hdma->cbase & ~0x3FF;
	index = (i % 7) * 4;
	hdma->chnum = (((hdma->base & 0xFF) - 8) / 20) & 0x1F;

	/* DMAꥯ */
#ifdef TADR_DMAMUX1_BASE
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB1ENR), RCC_AHB1ENR_DMAMUX1EN);

	if(hdma->Init.Direction == DMA_MEMORY_TO_MEMORY)
	    hdma->Init.Request = DMA_REQUEST_MEM2MEM;

	/* DMAMUX˥ڥեͥ */
	sil_wrw_mem((uint32_t *)(DMAmux_Channel(hdma)), hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID);
	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

	if(hdma->Init.Request > 0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR3){
		/* ɬפʤDMAMUXꥯͥ졼νԤ*/
		index = hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID;
		hdma->genreqno = index - 1;
		/* DMAMUXꥯȥͥ졼 쥸ꥻå */
		sil_wrw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), 0);
		/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
	}
	else
		hdma->genreqno = -1;

#else
	sil_modw_mem((uint32_t *)(hdma->base+TOFF_DMA_CSELR), (0xF << index), (hdma->Init.Request << index));

#endif

	/* DMA-CCR쥸 */
	tmp = sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR));

	/* DMA⡼ɥӥåȤƥݥꥣ˥ꥢ */
	tmp &= ((uint32_t)~(DMA_CCR_DIR  | DMA_CCR_MEM2MEM | DMA_CCR_PINC  | \
                        DMA_CCR_MINC | DMA_CCR_PSIZE   | DMA_CCR_MSIZE | \
                        DMA_CCR_CIRC | DMA_CCR_PL));

	/* DMA⡼ɤƥݥꥣ */
	tmp |=   hdma->Init.Direction        |
          hdma->Init.PeriphInc           | hdma->Init.MemInc           |
          hdma->Init.PeriphDataAlignment | hdma->Init.MemDataAlignment |
          hdma->Init.Mode                | hdma->Init.Priority;

	/* DMA⡼ */
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), tmp);

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  CHANNEL DMAλؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_deinit(DMA_Handle_t *hdma)
{
	int i, index;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < NUM_CHDMA ; i++){
		if(dma_base[i] == hdma->cbase)
			break;
	}
	if(i == NUM_CHDMA)
		return E_PAR;
	hdma->chid = i % 7;
	pDmaHandler[i] = NULL;
	index = (i % 7) * 4;

	/* DMA֥͡ */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), 0);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CNDTR), 0);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CPAR), 0);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CMAR), 0);

	/* ƥե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_ISR_GIF << index));
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTCIF << index));
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CHTIF << index));
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTEIF << index));

	/* DMAꥯȥꥢ */
#ifdef TADR_DMAMUX1_BASE
	sil_wrw_mem((uint32_t *)(DMAmux_Channel(hdma)), 0);

	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

	/* ɬפʤꥯȥͥ졼Υѥ᡼Ԥ */
	if(hdma->Init.Request >  0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR3){
		/* DMAMUXꥯȥͥ졼 쥸ꥻå */
		sil_wrw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), 0);
		/* DMAMUXꥯȥͥ졼 Сե饰򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
	}
#else
	sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMA_CSELR), (0xF << index));
#endif

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  STREAM DMAϴؿ
 *  parameter1  hdma:       DMAϥɥؤΥݥ
 *  parameter2  SrcAddress: ɥ쥹
 *  parameter3  DstAddress: ǥƥ͡󥢥ɥ쥹
 *  parameter4  DataLength: žĹ
 *  return ER
 */
ER
dma_start(DMA_Handle_t *hdma, uint32_t SrcAddress, uint32_t DstAddress, uint32_t DataLength)
{
	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);
	hdma->status = DMA_STATUS_BUSY;

#ifdef TADR_DMAMUX1_BASE
	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

	if(hdma->genreqno >= 0)	/* DMAMUXꥯȥͥ졼 Сե饰򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
#endif

	/* ǡĹ */
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CNDTR), DataLength);

	if((hdma->Init.Direction) == DMA_MEMORY_TO_PERIPH){
		/* ꤫ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CPAR), DstAddress);
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CMAR), SrcAddress);
	}
	else{	/* ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CPAR), SrcAddress);
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CMAR), DstAddress);
	}

	/* ߥ֥͡ */
	sil_orw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), (DMA_CCR_TCIE | DMA_CCR_HTIE | DMA_CCR_TEIE));

#ifdef TADR_DMAMUX1_BASE
    /* DMAMUX 󥯥ʥ */
	if((sil_rew_mem((uint32_t *)(DMAmux_Channel(hdma))) & DMAMUX_CxCR_SE) != 0){
		/* DMAMUX 󥯥ʥСߤ */
		sil_orw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
    }
	if(hdma->genreqno >= 0){
		/* DMAMUXꥯȥͥ졼  overrunߤͭ */
		sil_orw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
	}
#endif

	/* DMA */
	sil_orw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);

	return E_OK;
}

/*
 *  STREAM DMAߴؿ
 *  parameter1  hdma  : DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_end(DMA_Handle_t *hdma)
{
	uint32_t tick = 0;

	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);

	/* DMAԤ */
	while((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_EN) != 0){
	    /* Check for the Timeout */
		if(tick > DMA_TRS_TIMEOUT){
			/* ॢȥ顼 */
			hdma->ErrorCode |= DMA_ERROR_TIMEOUT;
			return E_TMOUT;
		}
    	dly_tsk(1);
	}
	hdma->status = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  STREAM DMA߽ؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 */
void
dma_inthandler(DMA_Handle_t *hdma)
{
	uint32_t index = hdma->chid * 4;

	/*
	 *  ž顼
	 */
	if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_ISR)) & (DMA_ISR_TEIF << index)) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_TEIE) != 0){
			/* ž顼ߥꥢ */
			sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_TEIE);
#ifdef TADR_DMAMUX1_BASE
			/* DMAMUX 󥯥Сߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
#endif
			/* ž顼ե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTEIF << index));
#ifdef TADR_DMAMUX1_BASE
			/* DMAMUX 󥯥Сե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

			if(hdma->genreqno >= 0){
				/* DMAMUXꥯȥͥ졼 overrunߤ̵ */
				sil_andw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
			/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
				sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
			}
#endif
			/* ž顼֤ */
			hdma->ErrorCode |= DMA_ERROR_TE;
			hdma->status = DMA_STATUS_READY_ERROR;
			/* 顼Хå */
			if(hdma->errorcallback != NULL)
				hdma->errorcallback(hdma);
		}
	}

	/*
	 *  ϡžλ
	 */
	if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_ISR)) & (DMA_ISR_HTIF << index)) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_HTIE) != 0){
			if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_CIRC) == 0){
				/* ϡžߤ򥯥ꥢ */
				sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_HTIE);
			}
#ifdef TADR_DMAMUX1_BASE
			/* DMAMUX 󥯥Сߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
#endif
			/* ϡžλե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CHTIF << index));
#ifdef TADR_DMAMUX1_BASE
			/* DMAMUX 󥯥Сե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

			if(hdma->genreqno >= 0){
				/* DMAMUXꥯȥͥ졼 overrunߤ̵ */
				sil_andw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
				/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
				sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
			}
#endif
			/* ϡžơѹ */
			hdma->status = DMA_STATUS_READY_HMEM0;
			/* ϡžλХå */
			if(hdma->xferhalfcallback != NULL)
				hdma->xferhalfcallback(hdma);
		}
	}

	/*
	 *  žλ
	 */
	if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_ISR)) & (DMA_ISR_TCIF << index)) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_TCIE) != 0){
			if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_CIRC) == 0){
				/* žߤ򥯥ꥢ */
				sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_TCIE);
			}
#ifdef TADR_DMAMUX1_BASE
			/* DMAMUX 󥯥Сߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
#endif
			/* žλե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTCIF << index));
#ifdef TADR_DMAMUX1_BASE
			/* DMAMUX 󥯥Сե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

			if(hdma->genreqno >= 0){
				/* DMAMUXꥯȥͥ졼 overrunߤ̵ */
				sil_andw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
				/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
				sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
			}
#endif
			/* ơλ */
			hdma->ErrorCode = DMA_ERROR_NONE;
			hdma->status = DMA_STATUS_READY_MEM0;
			/* žλХå */
			if(hdma->xfercallback != NULL)
				hdma->xfercallback(hdma);
		}
	}
}

/*
 *  STREAM DMA ߥӥ롼
 */
void
channel_dma_isr(intptr_t exinf)
{
	dma_inthandler(pDmaHandler[(uint32_t)exinf]);
}


/*
 *  WINDOW WATCH-DOGǡ¸ΰ
 */
static uint32_t mcount, mcounter, wdogcount;

/*
 *  WINDOW WATCH-DOG
 *  parameter1  port: WDOGݡֹ
 *  parameter2  hini: WDOGϥɥΥݥ
 *  return ｪλ E_OK
 */
ER
wdog_init(ID portid, WDOG_Init_t *hini)
{
	if(hini == NULL || portid != WDOG1_PORTID)
		return E_PAR;
	if(hini->Window > 127 || hini->Window < 64)
		return E_PAR;
	if(hini->Counter > 127 || hini->Counter < 64)
		return E_PAR;

	mcount    = hini->MCounter;
	mcounter  = 0;
	wdogcount = hini->Counter;
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_WWDGEN);
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	if(mcount == 0)
		sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CFR), (WWDG_CFR_WDGTB | hini->Window));
	else
		sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CFR), (WWDG_CFR_EWI | WWDG_CFR_WDGTB | hini->Window));
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), (WWDG_CR_WDGA | wdogcount));
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGλ
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_deinit(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_WWDGEN);
	wdogcount = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGꥻå
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_reset(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	mcounter = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGߥϥɥ
 */
void
wdog_handler(void)
{
	uint32_t base = TADR_WWDG_BASE;

	if((sil_rew_mem((uint32_t *)(base+TOFF_WWDG_CFR)) & WWDG_CFR_EWI) != 0
			&& (sil_rew_mem((uint32_t *)(base+TOFF_WWDG_SR)) & WWDG_SR_EWIF) != 0){
		sil_andw_mem((uint32_t *)(base+TOFF_WWDG_SR), WWDG_SR_EWIF);
		if(mcounter++ < mcount)
			sil_wrw_mem((uint32_t *)(base+TOFF_WWDG_CR), wdogcount);
	}
}


/*
 *  LED³ݡ
 *
 *  ĥI/OܡɤLED1-3ϥץޥ֥ϥݡ0
 *  ˥С𤷤³Ƥ롥
 */

typedef struct gio_confstruct
{
	uint32_t    clkbit;
	uint32_t    giobase;
	uint32_t    pinpos;
	uint32_t    pinpp;
} gio_conftype;

static const gio_conftype led_confdata[] = {
#if defined(TOPPERS_STM32L476_DISCOVERY)
	{ RCC_AHB2ENR_GPIOBEN, TADR_GPIOB_BASE, PINPOSITION2, GPIO_NOPULL},
	{ RCC_AHB2ENR_GPIOEEN, TADR_GPIOE_BASE, PINPOSITION8, GPIO_NOPULL}
#elif defined(TOPPERS_STM32L4R5_NUCLEO144)
	{ RCC_AHB2ENR_GPIOCEN, TADR_GPIOC_BASE, PINPOSITION7, GPIO_NOPULL},
	{ RCC_AHB2ENR_GPIOBEN, TADR_GPIOB_BASE, PINPOSITION7, GPIO_NOPULL},
	{ RCC_AHB2ENR_GPIOBEN, TADR_GPIOB_BASE, PINPOSITION14, GPIO_NOPULL}
#else
	{ RCC_AHB2ENR_GPIOAEN, TADR_GPIOA_BASE, PINPOSITION5, GPIO_PULLUP}
#endif
};
#define NUM_LED   (sizeof(led_confdata)/sizeof(gio_conftype))


/*
 *  LED³ݡȽ
 */ 
void
led_init(intptr_t exinf)
{
	const gio_conftype *pled = &led_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t i;

	for(i = 0 ; i < NUM_LED ; pled++, i++){
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), pled->clkbit);
		/* ⡼ */
		GPIO_Init_Data.mode      = GPIO_MODE_OUTPUT;
		/* ץ륢å ץ */
		GPIO_Init_Data.pull      = pled->pinpp;
		/* ϥ⡼ */
		GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
        /* ԡ */
		GPIO_Init_Data.speed     = GPIO_SPEED_HIGH;
		gpio_setup(pled->giobase, &GPIO_Init_Data, pled->pinpos);

		/*  */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (1 << pled->pinpos)<<16);
	}
}

/*
 *  LED³ݡɤ߽Ф
 */
uint_t
led_in(void)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t data, i;

	for(i = 0, data = 0 ; i < NUM_LED ; pled++, i++){
		if((sil_rew_mem((uint32_t *)(pled->giobase+TOFF_GPIO_IDR)) & (1<<(pled->pinpos))) != 0)
			data |= (1<<i);
	}
	return data;
}

/*
 *  LED³ݡȽ񤭹
 */ 
void
led_out(unsigned int led_data)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t reg1, reg2, i;

	/* ͤLED³ӥåȰʳѹʤ */
	for(i = 0 ; i < NUM_LED ; pled++, i++){
		reg1 = reg2 = 0;
		if((led_data & (1<<i)) != 0)
			reg2 = 1 << pled->pinpos;
		else
			reg1 = 1 << pled->pinpos;

		/* 񤭹ߥǡƽ񤭹 */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (reg1<<16) | reg2);
	}
}

/*
 * LEDȥåθꡦɤ߹ߴؿ
 */

/*
 *  LEDξ¸ѿ
 */	
unsigned int LedState;


/*
 *  LED
 */
void 
set_led_state(unsigned int led, unsigned char state){
	if (state == ON) {
		LedState = LedState | led;
	} else {
		LedState = LedState & ~led;
	}
	led_out(LedState);
}


static const gio_conftype sw_confdata[] = {
#if defined(TOPPERS_STM32L476_DISCOVERY)
	{ RCC_AHB2ENR_GPIOAEN, TADR_GPIOA_BASE, PINPOSITION0, GPIO_PULLDOWN},
	{ RCC_AHB2ENR_GPIOAEN, TADR_GPIOA_BASE, PINPOSITION1, GPIO_PULLDOWN},
	{ RCC_AHB2ENR_GPIOAEN, TADR_GPIOA_BASE, PINPOSITION2, GPIO_PULLDOWN},
	{ RCC_AHB2ENR_GPIOAEN, TADR_GPIOA_BASE, PINPOSITION3, GPIO_PULLDOWN},
	{ RCC_AHB2ENR_GPIOAEN, TADR_GPIOA_BASE, PINPOSITION5, GPIO_PULLDOWN}
#else
	{ RCC_AHB2ENR_GPIOCEN, TADR_GPIOC_BASE, PINPOSITION13, GPIO_NOPULL}
#endif
};
#define NUM_SW    (sizeof(sw_confdata)/sizeof(gio_conftype))

static void
sw_int(void)
{
	if(sw_func != NULL)
		sw_func(0);
}

/*
 * PUSHå³ݡȽ
 */
void
switch_push_init(intptr_t exinf)
{
	const gio_conftype *psw = &sw_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t i;

	for(i = 0 ; i < NUM_SW ; psw++, i++){
		/* ϥݡȤ */
		/* Enable the BUTTON Clock */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), psw->clkbit);
		/* ⡼ */
		GPIO_Init_Data.mode  = GPIO_MODE_IT_FALLING;
		/* ץ륢å ץ */
		GPIO_Init_Data.pull  = psw->pinpp;
        /* ԡ */
		GPIO_Init_Data.speed = GPIO_SPEED_HIGH;
		gpio_setup(psw->giobase, &GPIO_Init_Data, psw->pinpos);
		if(psw->pinpos >= EXTI10_BASENO)
			exti10_func[psw->pinpos - EXTI10_BASENO] = sw_int;
		else if(psw->pinpos >= EXTI5_BASENO)
			exti5_func[psw->pinpos - EXTI5_BASENO] = sw_int;
	}
}

/*
 *  PUSHåХåؿ
 */
void setup_sw_func(intptr_t exinf)
{
	sw_func = (void (*)(int))exinf;
}

/*
 *  PUSHåӥ
 */
void
sw_isr(intptr_t exinf)
{
	uint32_t istatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1));
	if(sw_func != NULL)
		sw_func((int)exinf);
    sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1), istatus);
}

/*
 * EXTI5ߥϥɥ
 */
void
exti5_handler(void)
{
	uint32_t istatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1));
	uint32_t i;
	for(i = 0 ; i < NUM_EXTI5_FUNC ; i++){
		if((istatus & (1<<(i+EXTI5_BASENO))) != 0){
			if(exti5_func[i] != NULL)
				(exti5_func[i])();
		}
	}
    sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1), istatus);
}

/*
 * EXTI10ߥϥɥ
 */
void
exti10_handler(void)
{
	uint32_t istatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1));
	uint32_t i;
	for(i = 0 ; i < NUM_EXTI10_FUNC ; i++){
		if((istatus & (1<<(i+EXTI10_BASENO))) != 0){
			if(exti10_func[i] != NULL)
				(exti10_func[i])();
		}
	}
    sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1), istatus);
}


/*
 * DIPSWμФ
 */
uint_t
switch_dip_sense(void)
{
	return dipsw_value;
}

